/*
 * Copyright (c) 2026 Valve Corporation
 * SPDX-License-Identifier: MIT
 */

#ifndef FD_HW_COMMON_H
#define FD_HW_COMMON_H

#include <stdbool.h>
#include <stdint.h>

#include "util/half_float.h"
#include "util/u_math.h"

#include "adreno_pm4.xml.h"
#include "adreno_common.xml.h"

#endif /* FD_HW_COMMON_H */
